/*
 *  client/functions-conf.c
 *		The functions to actually execute the commands
 *		for the client.  Commands for handling
 *		conferences.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */

#include <config.h>
#include "conf.h"

#include <stdio.h>
/* #include <time.h> included from kom-types.h */
/* #include <malloc.h> */
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef __vax__
#define __SIZE_T included from string.h
#endif
#include <kom-types.h> /* Before zmalloc.h for those that have size_t in 
			  sys/types.h */
#include <zmalloc.h>
#include <s-string.h>
#include <s-collat.h>

#include <kom-errno.h>
#include <services.h>

/* Must be included before "read-line.h" on a vax because it includes <sgtty.h>
 * and that does not work after inclusion of <sys/ioctl.h> because of
 * stupid include-file construction.
 */
#include "xoutput.h"

#include "copy.h"
#include "read-line.h"
#include "commands.h"
#include "parser.h"
#include "texts.h"
#include "error.h"
#include "cache.h"
#include "quit.h"

#include "internal.h"
#include "cmds.h"

#include "defaults.h"

/* This is to emphasize that a function is exported. */
#define Export

/* Forward declaration */
static Success
do_leave_conf (Conf_no	conf_to_leave);



/*
 *  The one and only  parameter to all the external functions, is a
 *  String containing the rest of the line, following the command.
 *  E g if the user gave the command "Skapa m|te Inl{gg }t mig",
 *  then the String "Inl{gg }t mig" is passed as the parameter
 *  'ARGUMENT' to the function responsible for the command 'skapa
 *  m|te'.  Leading and trailing spaces are skipped.  EMPTY_STRING
 *  is passed if no arguments to the command was given.
 */




/*  Create a new conference  */
Export  Success
cmd_create_new_conference (String	new_conf_name)
{
#define FUNCTION "create_new_conference()"

    volatile Conf_type	  type_of_conf;
    Bool		  original;
    Bool		  open;
    Bool		  closed;
    Bool		  secret;
    Bool		  write_protect;
    String		  input			= EMPTY_STRING;
    Conf_no		  new_conf_no;
    Parse_token		* words;
    Parse_token *volatile all_words		= NULL;
    Conf_z_info_list	  confs;
    Success		  retval;


#define CLEAN_UP()						\
    {								\
	s_clear (&new_conf_name);   				\
	s_clear (&input);					\
	free_tokens (all_words);				\
    }


    TOPLOOP_SETJMP();

    if (s_empty (new_conf_name))
    {
	xprintf ("%s?\n= ", txts.Whats_new_conf_name);
	read_line (&new_conf_name);
	new_conf_name = s_strip_trailing (new_conf_name,
					  command_separators);
	if (s_empty (new_conf_name))
	{
	    RETURN (FAILURE);
	}
    }
    else
    {
	s_puts (new_conf_name);
    }
    newline();
    newline();

    /* Get the type of the conference (open, write protect, original, ...) */
    xprintf ("%s? %s:\n%s, %s, %s, %s, %s\n= ",
	    txts.What_type_of_conf,
	    txts.Choose_one_or_more_of,
	    txts.open, txts.closed, txts.original,
	    txts.secret, txts.write_protected);
    read_line (&input);		/* BUG! Retval not checked! */

    words = tokenize (input, command_separators);
    if (words == NULL)
    {
	fatal1 (CLIENT_OUT_OF_MEMORY, "tokenize() failed");
    }

    /* Empty string? Exit then. */
    if (s_empty (words->word))
    {
	RETURN (FAILURE);
    }

    all_words = words;		/* Save the start of the list */
    open = closed = original = secret = write_protect = FALSE;
    while (! s_empty (words->word))
    {
	/* DIRTY!!! */
	if (s_usr_strhead (words->word, s_fcrea_str (txts.open),
			   DEFAULT_COLLAT_TAB))
	{
	    if (closed)
	    {
		xprintf ("%s.\n", txts.Cant_be_open_and_closed);
		RETURN (FAILURE);
	    }
	    open = TRUE;
	}
	else if (s_usr_strhead (words->word, s_fcrea_str (txts.closed),
				DEFAULT_COLLAT_TAB))
	{
	    if (open)
	    {
		xprintf ("%s.\n", txts.Cant_be_open_and_closed);
		RETURN (FAILURE);
	    }
	    closed = TRUE;
	}
	else if (s_usr_strhead (words->word, s_fcrea_str (txts.original),
				DEFAULT_COLLAT_TAB))
	    original = TRUE;
	else if (s_usr_strhead (words->word, s_fcrea_str (txts.secret),
				DEFAULT_COLLAT_TAB))
	    secret = TRUE;
	else if (s_usr_strhead (words->word,
				s_fcrea_str (txts.write_protected),
				DEFAULT_COLLAT_TAB))
	    write_protect = TRUE;
	else
	{
	    xprintf ("%s: ", txts.Not_understood_by_KOM);
	    s_puts (words->word);
	    newline();
	    RETURN (FAILURE);
	}

	words++;
    }	/* While list isn't empty */

    type_of_conf.rd_prot = ! open;
    type_of_conf.original = original;
    type_of_conf.secret = secret;
    type_of_conf.letter_box = FALSE;

    /* Okay, then let's try to create the conference */
    new_conf_no = kom_create_conf (new_conf_name, type_of_conf);
    if (new_conf_no == 0)
    {
	switch (kom_errno)
	{
	case  KOM_PERM:
	    xprintf ("%s\n", txts.Not_allowed_create_confs);
	    break;

	case  KOM_BAD_NAME:
	case  KOM_LONG_STR:
	    xprintf ("%s.\n", txts.Name_was_bad);
	    break;

	case  KOM_CONF_EXISTS:
	    xprintf ("%s.\n", txts.Name_occupied);
	    break;

	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR,
		   "create_conf() failed: %d", kom_errno);
	    break;
	}

	RETURN (FAILURE);
    }	/* If kom_create_conf() failed */
    else 
    {
	/* Join the new conference. */
	do_join_conf(new_conf_no);
    }

    /* Now, let's see if we need to set a super-conference */
    if (original || write_protect)
    {
	xprintf ("%s? ", txts.Which_conf_super_conf);
	read_line (&input);
	input = s_strip_trailing (input, command_separators);
	if (s_empty (input))
	{
	    RETURN (FAILURE);
	}

	confs = find_conf_or_pers_no (input);
	if (confs.no_of_confs != 1)
	{
	    inform_user (ambigous_or_no_conf, &confs);
	    RETURN (FAILURE);
	}

	retval = kom_set_super_conf (new_conf_no, confs.confs[0].conf_no);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_NO_CONNECT:
		fatal2 (CLIENT_CONNECTION_LOST, "set_super_conf() failed: %d",
		       confs.confs[0].conf_no);
		break;

		/* Hmm, what reasons are there for this failing?
		 * I can't see any.  If we aren't allowed to set
		 * it as super-conf, then we shouldn't be able to
		 * get its name either...		*/
	    default:
		fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR,
		       "set_super_conf() failed: %d", confs.confs[0].conf_no);
		break;
	    }
	}   /* If kom_set_super_conf() failed */

	/* And set the current person as the only one allowed
	 * to submit texts to the new conference.	    */
	retval = kom_set_permitted_submitters (new_conf_no,
					       (Conf_no) cpn);
	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_NO_CONNECT:
		fatal3 (CLIENT_CONNECTION_LOST,
			"set_permitted_submitters(%d tp be a submitter of %d) failed", 
			cpn, new_conf_no);
		break;

		/* Hmm, what reasons are there for this failing?
		 * I can't see any.  If we aren't allowed to set
		 * it as super-conf, then we shouldn't be able to
		 * get its name either...		*/
	    default:
		fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR,
			"set_permitted_submitters() failed",
			confs.confs[0].conf_no);
		break;
	    }
	}   /* If 'set_permitted_submitters()' failed */
    }   /* If super-conferece needed */

    RETURN (OK);

#undef	CLEAN_UP
#undef  FUNCTION
}   /* END: create_new_conference */


/* Remove a conference */
Export  Success
cmd_remove_conference (String conference)
{
#define FUNCTION "remove_conference()"

    Conf_z_info      whichconf = EMPTY_CONF_Z_INFO;

#define CLEAN_UP()				\
    {						\
	s_clear(&conference);			\
    }

    TOPLOOP_SETJMP();

    whichconf = ask_for_conf(&conference,
			     txts.What_conf_to_remove);
    if(whichconf.conf_no == 0)
    {
	RETURN (FAILURE);
    }

    if(whichconf.conf_no == cpn)
    {
	xprintf("%s",txts.Cannot_remove_oneself);
	RETURN (FAILURE);
    }

    if (kom_delete_conf(whichconf.conf_no) != FAILURE)
    {
	xprintf("%s ",txts.Successfully_removed);
	print_conf_name(whichconf.conf_no);
    }
    else
    {
	switch (kom_errno)
	{
	case  KOM_PERM:
	    xprintf("%s ",txts.Not_allowed_to_remove);
	    print_conf_name(whichconf.conf_no);
	    break;

	default:
	    fatal3 (CLIENT_UNEXPECTED_SERVER_ERROR,
		    "delete_conf(%d) error %d",
		    whichconf.conf_no, kom_errno);
	    break;
	}
    }

    RETURN (OK);
#undef  CLEAN_UP
#undef  FUNCTION
}

/* Find all names matching the string given as argument */
Export  Success
cmd_show_matching_names (String	name)
{
#define FUNCTION "show_matching_names()"

    Success		  retval;
    Conf_z_info_list	  matches	= EMPTY_CONF_Z_INFO_LIST;
    unsigned long	  i;


#define CLEAN_UP()					\
    {							\
	s_clear (&name);				\
	release_conf_z_info_list (&matches);		\
    }


    TOPLOOP_SETJMP();

    if (s_empty (name))
    {
	xprintf ("\n%s\n- ", txts.Give_name_to_match_for);
	read_line (&name);
    }
    else
	s_puts (name);

    newline();

    retval = kom_lookup_z_name (name, TRUE, TRUE, &matches);
    if (retval == FAILURE)
    {
	switch (kom_errno)
	{
	case  KOM_OUT_OF_MEMORY:
	    fatal1 (CLIENT_OUT_OF_MEMORY, "lookup_z_name() failed");
	    break;

	case  KOM_NO_CONNECT:
	    fatal1 (CLIENT_CONNECTION_LOST, "lookup_z_name() failed");
	    break;

	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR, 
		    "lookup_z_name() failed: %d",
		    kom_errno);
	    break;
	}
    }

    for ( i = 0 ;  i < matches.no_of_confs ;  i++ )
    {
	if (matches.confs[i].type.letter_box)
	{
	    xprintf("%s %u:\t\t%S\n", txts.Person,
		    matches.confs[i].conf_no,
		    matches.confs[i].name);
	}
	else
	{
	    xprintf("%s %u:\t\t%S\n", txts.Conference,
		    matches.confs[i].conf_no,
		    matches.confs[i].name);
	}
    }

    newline();

    RETURN (OK);

#undef CLEAN_UP
#undef FUNCTION
}   /* END: show_matching_names */


/*  Go to the specified conference.  Join it if not a member.  */
Export  Success
cmd_goto_named_conference (String   conf_name)
{
    /* static char	* this_function	= "goto_named_conference()"; */

    Conf_z_info		  go_to_conf = EMPTY_CONF_Z_INFO;
    Conf_no		  conf_to_go_to;
    Conference          * conf_info;
    Membership		* memb_ptr		= NULL;
    volatile Success	  retval		= FAILURE;
    Kom_err               error;

#define CLEAN_UP()				\
    {						\
	s_clear (&conf_name);			\
	zfree (memb_ptr);			\
	release_conf_z_info(&go_to_conf);	\
    }


    TOPLOOP_SETJMP();

    go_to_conf = ask_for_conf(& conf_name,
			      txts.What_conf_to_go_to);
    if (go_to_conf.conf_no == 0)
    {
	RETURN (FAILURE);
    }

    conf_to_go_to = go_to_conf.conf_no;
    xprintf("%S", go_to_conf.name);

    /* Well, is the user member of the conference? */
    if (locate_membership (conf_to_go_to, cpn, NULL) == -1)
    {
        conf_info = conf_stat (conf_to_go_to, &error);
            /* Should check error value of conf_stat */
        if (conf_info->presentation)
            show_text_no (conf_info->presentation);
	newline();
	/* Nope. Ask if s/he wants to join */
	xprintf ("%s. %s?\n",
		txts.Youre_not_member,
		txts.Want_to_join);
	if (yes_or_no_p())
	    retval = do_join_conf (conf_to_go_to);
	else
	{
	    RETURN (FAILURE);
	}

	xprintf("%S", go_to_conf.name);
    }

    review_reset();

    {
	Membership	  ship;
	Conference 	* conf;
	int 		  unread;

	locate_membership (conf_to_go_to, cpn, &ship);
	conf = conf_stat (conf_to_go_to, &error);
	unread = conf->first_local_no + conf->no_of_texts - 1 - 
	    (ship.last_text_read + ship.no_of_read);

	if (unread > 0)
	    xprintf(" - %d %s.\n", unread,
		    (unread == 1 ? txts.unread_singular
		               : txts.unread_plural));
	else
	    xprintf(" - %s.\n", txts.no_unread);
	
	goto_conference (conf_to_go_to);
	release_conf(conf);
	zfree(conf);
    }

    RETURN (retval);

#undef  CLEAN_UP
}   /* END: goto_named_conference */




/*  Join (become a member of) a conference.  */
Export  Success
cmd_join_conference (String   conf_name)
{
/*    char		* this_function	= "join_conference()"; */

    Conf_z_info		  join_conf = EMPTY_CONF_Z_INFO;
    Conf_no		  conf_to_join;


#define CLEAN_UP()				\
    {						\
	s_clear (&conf_name);			\
	release_conf_z_info(&join_conf);	\
    }


    TOPLOOP_SETJMP();

    join_conf = ask_for_conf(& conf_name,
			     txts.What_conf_to_join);
    if (join_conf.conf_no == 0)
    {
	RETURN (FAILURE);
    }
    conf_to_join = join_conf.conf_no;

    if (do_join_conf (conf_to_join) == OK)
    {
	goto_conference (conf_to_join);
	RETURN (OK);
    }
    else
    {
	RETURN (FAILURE);
    }

    /*NOTREACHED*/ /* Thanks apogee that found this one! */
    RETURN (OK);
#undef  CLEAN_UP
}   /* END: join_conference */



/*  Leave (resign as a member of) a conference.  */
Export  Success
cmd_leave_conference (String   conf_name)
{
#define FUNCTION "cmd_leave_conference()"

    Conf_z_info		  leave_conf = EMPTY_CONF_Z_INFO;
    Conf_no		  conf_to_leave;

#define CLEAN_UP()				\
    {						\
	s_clear (&conf_name);			\
	release_conf_z_info(&leave_conf);	\
    }


    TOPLOOP_SETJMP();

    leave_conf = ask_for_conf(& conf_name,
			     txts.What_conf_to_leave);
    if (leave_conf.conf_no == 0)
    {
	RETURN (FAILURE);
    }
    conf_to_leave = leave_conf.conf_no;

    if (do_leave_conf (conf_to_leave) == OK)
    {
	if (conf_to_leave == ccn)
	    goto_conference (0); /* Leaving conf. Going nowhere. */
	RETURN (OK);
    }
    else
    {
	RETURN (FAILURE);
    }

    /*NOTREACHED*/
    RETURN (OK);
#undef  CLEAN_UP
#undef FUNCTION
}   /* END: leave_conference */



/* Var {r jag
 * Prints some information about the name, conference and conference system. */
Export  Success
cmd_where_am_i (String	argument)

{
    extern char		* server_name;
    extern Info		  serv_info;
    extern char		* client_name;

#define CLEAN_UP()				\
    {						\
	s_clear (&argument);			\
    }

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    newline();
    newline();

    print_conf_name (cpn);
    
    if (ccn == 0)
    {
	xprintf (txts.is_not_in_any_conf);
    }
    else	
    {
	Membership	  ship = EMPTY_MEMBERSHIP;
	Kom_err		  error;
	int		  unread;

	locate_membership(ccn, cpn, &ship);

	xprintf (txts.is_in_conf);
	print_conf_name (ccn);
	unread = conf_last(ccn, &error)
	    - ( ship.last_text_read + ship.no_of_read );
	if (unread == 1)
	    xprintf (txts.paranthetisised_unread, unread);
	else
	    xprintf (txts.paranthetisised_unreads, unread);
    }
    xputchar ('.');
    newline();

    xprintf (txts.connected_to, server_name, serv_info.version, 
	     client_name, VERSION);

    RETURN (OK);

#undef CLEAN_UP
}   /* END: where_am_i */


/*
 *  Show the status for a person or a conference.
 */
Export  Success
cmd_status (String	  name)

{
    Conf_z_info		  conf_to_stat = EMPTY_CONF_Z_INFO;
    Conference	*volatile conf		= NULL;
    Member_list	*volatile members	= NULL;
    Person	*volatile pers		= NULL;
    Kom_err		  error_c, error_p;

#define CLEAN_UP()				\
    {						\
	s_clear (&name);			\
	zfree (release_conf (conf));		\
	zfree (release_member_list (members));	\
	zfree (release_person (pers));		\
        release_conf_z_info(&conf_to_stat);	\
    }

    TOPLOOP_SETJMP();

    conf_to_stat = ask_for_conf (&name, txts.What_to_get_stat_for);
    if (conf_to_stat.conf_no == 0)
    {
	RETURN (FAILURE);
    }
    newline();

    conf = conf_stat (conf_to_stat.conf_no, &error_c);
    if (conf == NULL)
    {
	inform_user (ambigous_or_no_conf, NULL);
	RETURN (FAILURE);
    }
    members = member_list (conf_to_stat.conf_no, &error_c);
    if (members == NULL)
    {
	inform_user (ambigous_or_no_conf, NULL);
	RETURN (FAILURE);
    }

    if (conf_to_stat.type.letter_box)
    {
	pers = pers_stat (conf_to_stat.conf_no, &error_p);
	if (pers == NULL)
	{
	    inform_user (ambigous_or_no_conf, NULL);
	    RETURN (FAILURE);
	}

	show_short_conf_status (conf_to_stat, conf);
	newline();
	show_pers_status (conf_to_stat, conf, pers);
	newline();
        if (conf->msg_of_day)
        {
            xprintf ("%s %s:\n", s_crea_c_str (conf->name), txts.Has_note);
            show_text_no (conf->msg_of_day);
        }
	show_more_conf_status (conf_to_stat, conf);
	newline();
/*	show_membership_list (conf_to_stat, conf, pers, memberships); */
    }
    else
    {
	show_short_conf_status (conf_to_stat, conf);
	newline();
        if (conf->msg_of_day)
        {
            xprintf ("\n%s %s:\n", s_crea_c_str (conf->name), txts.Has_note);
            show_text_no (conf->msg_of_day);
            newline();
            newline();
        }
	show_more_conf_status (conf_to_stat, conf);
	newline();
	if (members)
	    show_member_list (conf_to_stat, conf, members);
	else
	    xprintf ("%s\n", txts.Conf_has_no_members);
    }
    
    RETURN (OK);

#undef CLEAN_UP
}   /* END: cmd_status() */



/* ==================================================================== */
/*		Internal functions for 'functions-*.c'			*/




/*
 *  Find all conferences matching the name NAME.  Also supports the
 *  notation "conference nnn" and "person nnn", where nnn is a number.
 *  If one of the latter notations are used, the 'pers_or_conf' field
 *  of the result will reflect the notation used, if it *is* a person.
 *  If the notation 'person nnn' is used and it *isn't* a person, then
 *  it won't match, i e the EMPTY_CONF_Z_INFO_LIST will be returned.
 */
Export  Conf_z_info_list
find_conf_or_pers_no (String	  name)

{
#define FUNCTION "find_conf_or_pers_no()"

    Conf_no		  conf;
    String_size		  error_check; /* To check for errors */
    String_size		  start_pos;
    String		  first_word;
    Conference	*volatile conf_s	= NULL;
    Conf_z_info_list	  result	= EMPTY_CONF_Z_INFO_LIST;
    Kom_err		  error;
    Success		  retval;

#define CLEAN_UP()				\
    {						\
	zfree (conf_s);				\
    }

    TOPLOOP_SETJMP();

    /* Strip trailing blanks for the ease of parsing "Person nnn"
     * and "Conference nnn"					*/
    name = s_strip_trailing (name, command_separators);

    /* Is it of the format "Person nnn"? */
    /* Well, let's try and see if we have any luck	*/
    start_pos = 0;
    first_word = s_strtok (name, &start_pos, command_separators);
    /* I don't particularly _like_ this, but... */
    if (  (start_pos < s_strlen(name))	/* More than one word? */)
    {
	int is_person = FALSE;

	if ((((s_usr_strhead (first_word, s_fcrea_str (txts.Person),
			      DEFAULT_COLLAT_TAB))
	      && ((is_person = TRUE)))
	     || (s_usr_strhead (first_word, s_fcrea_str (txts.Conference),
				DEFAULT_COLLAT_TAB)))
	    && (conf = (Conf_no) s_strtol (s_fsubstr (name, start_pos,
						      END_OF_STRING),
					   &error_check,
					   DEFAULT_NUMBER_BASE))
	    && (error_check == s_strlen(name) - start_pos)
	    && (conf_s = conf_stat (conf, &error)))
	{
	    result.confs = zmalloc (sizeof (Conf_z_info));
	    if (result.confs == NULL)
	    {
		/* Seem to be running out of memory */
		fatal1 (CLIENT_OUT_OF_MEMORY, "");
	    }
	    result.no_of_confs = 1;
	    result.confs[0].conf_no = conf;
	    result.confs[0].name = conf_s->name;
	    result.confs[0].type = conf_s->type;
	    if (!is_person)
	    {
		result.confs[0].type.letter_box = FALSE;
	    }
	    RETURN(result);
	}
    }

    /* It was neither of the form "Person nnn" nor "Conference nnn" */
    {
	/* Then assume it is a name and find it's number */
	retval = kom_lookup_z_name (name, TRUE, TRUE, &result);
	if (retval == FAILURE)
	{
	    /* Waaahhh!!  I just hate this! */
	    fatal2 (CLIENT_ERROR_TOO_boring,
		    "kom_lookup_name() failed: %s", s_crea_c_str(name));
	}

    }

    /* Phew! Finished at last. */
    RETURN (result);

#undef  CLEAN_UP
#undef  FUNCTION
}   /* END: find_conf_or_pers_no */



/*
 *  Ask the user for the name or number of one conference or person.
 */
Export  Conf_z_info
ask_for_conf (String			* name,
	      const unsigned char	* prompt)
{
    Conf_z_info_list	  matching	= EMPTY_CONF_Z_INFO_LIST;
    Conf_z_info		  result	= EMPTY_CONF_Z_INFO;

#define CLEAN_UP()				\
    {						\
	release_conf_z_info_list (&matching);	\
    }

    TOPLOOP_SETJMP();

    if (s_empty (*name))
    {
	xprintf ("\n%s - ", prompt);
	read_line (name);
	newline();
	*name = s_strip_trailing (*name, command_separators);
	if (s_empty (*name))
	{
	    RETURN (result);
	}
    }

    matching = find_conf_or_pers_no (*name);
    if (matching.no_of_confs != 1)
    {
	inform_user (ambigous_or_no_conf, &matching);
    }
    else
    {
	copy_conf_z_info(&result, &matching.confs[0]);
	xprintf ("%S\n", result.name);
    }

    RETURN (result);

#undef CLEAN_UP
}   /* END: ask_for_conf() */



/*
 * Set the new current conference to NEW_CONF.
 * NEW_CONF can be 0 which means, dont go to any conference.
 */
Export  Success
goto_conference (Conf_no	new_conf)

{
#define FUNCTION "goto_conference()"
    Conf_no		  i;		/* Loop variable */
    Conf_no		  retval;
    Membership		  membership = EMPTY_MEMBERSHIP;
    Kom_err		  error;

#define CLEAN_UP()				\
    {						\
    }

    TOPLOOP_SETJMP();

#if 0
// *      /* Save what texts we have read in the old conference */
// *      /* BUG! retval not checked! */
// *      kom_mark_as_read (ccn,
// *  		      internal_state.member_info.no_of_read,
// *  		      internal_state.member_info.read_texts);
#endif

    retval = cache_change_conf (new_conf, &error);
    if (retval == 0)
    {
	switch (error)
	{
	case KOM_NO_ERROR:
	    /* This is no error. 
	       We are just going to the nowhere conference. */
	    break;

	case KOM_UNDEF_CONF:
	    xprintf ("%s\n", txts.No_conf_with_that_name);
	    break;

 	case KOM_NOT_MEMBER:
	    xprintf ("%s\n", txts.Youre_not_member);
	    break;

	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR, "cache_change_conf(): %d",
		   new_conf);
	}

	RETURN (FAILURE);
    }
    /* ////////////////////////////// */


#if 0
// *      retval = kom_query_read_texts (cpn, new_conf,
// *  				   &internal_state.member_info);
    if (retval == FAILURE)
    {
	/* I told you this was boring! */
	fatal2 (CLIENT_ERROR_TOO_boring,
	       "kom_query_read_texts() failed: %d", new_conf);
    }
#endif
#if 0
// *      internal_state.size_of_read_list =
// *  		internal_state.member_info.no_of_read;
// *  
// *      retval = person_stat (cpn,	/* //////////////// */ ); /* What do I need this for? */
// *      if (retval == FAILURE)
// *      {
// *  	/* I hate it! */
// *  	fatal (CLIENT_ERROR_TOO_boring,
// *  			    this_function,
// *  			    "kom_get_person_stat() failed",
// *  			    (fatal_info) cpn);
// *      }
// *
#endif

    if (new_conf != 0) 
    {
	/* Find the position in the member list of the new conference */
	i = locate_membership (new_conf, cpn, &membership);
	reading_state.local_last_normally_read = membership.last_text_read;
    } 

    RETURN (OK);

#undef CLEAN_UP
#undef FUNCTION
}   /* END: goto_conference */



/*
 *  Perform the actions to join a conference.  Inform the
 *  user if he is not allowed to join it.
 */
Export  Success
do_join_conf (Conf_no	conf_to_join)

{
#define FUNCTION "do_join_conf()"

    Success		  retval;
    Conference	*volatile conf			= NULL;
    Conference		* supervisor_conf	= NULL;
    Kom_err		  error;
    Person	*volatile pers_ptr		= NULL;


#define CLEAN_UP()					\
    {							\
	zfree (release_person (pers_ptr));		\
	zfree (release_conf (conf));			\
	zfree (release_conf (supervisor_conf));		\
    }

    TOPLOOP_SETJMP();


    pers_ptr = pers_stat (cpn, &error);
    if (pers_ptr == NULL)
	fatal2 (CLIENT_SERVER_ERROR,
		"Can't get stat for CPN: %d", error);
    /* BUG: Should be cache_add_member */
    retval = kom_add_member (conf_to_join, cpn, DEFAULT_CONF_PRIORITY,
			     pers_ptr->no_of_confs);
    if (retval == FAILURE)
    {
	switch (kom_errno)
	{
	case  KOM_ACCESS:
	    xprintf ("%s. %s\n%s ",
		    txts.Conf_is_closed,
		    txts.For_membership_contact_supervisor,
		    txts.Supervisor_is);
	    conf = conf_stat (conf_to_join, &error);
	    if (conf == NULL)
	    {
		fatal2 (CLIENT_ERROR_TOO_boring,
			"get_conf_stat() failed: %d", conf_to_join);
	    }

	    print_conf_name (conf->supervisor);
	    newline();
	    RETURN (FAILURE);
	    /*NOTREACHED*/
	    break;

	case  KOM_NO_CONNECT:
	    fatal1 (CLIENT_CONNECTION_LOST, "add_member() failed");
	    break;

	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR,
		    "add_member() failed: %d", conf_to_join);
	}
    }

    /* Invalidate the membership-cache. */
    first_membership(0,NULL);

    RETURN (OK);

#undef CLEAN_UP
#undef FUNCTION
}   /* END: do_join_conf */


/*
 * Perform the actions to leave a conference. Inform the user if he was not
 * a member.
 */
static Success
do_leave_conf (Conf_no	conf_to_leave)
{
#define FUNCTION "do_leave_conf()"

    Success		  retval;

#define CLEAN_UP()				\
    {						\
    }

    TOPLOOP_SETJMP();

    if (locate_membership (conf_to_leave, cpn, NULL) == -1) {
	/* We were not members. Can't the user keep track of what conferences
	   he is a member of? Do nothing. */
	xprintf ( txts.You_were_not_member_on_leave );
	RETURN (FAILURE);
    }
    retval = kom_sub_member(conf_to_leave, cpn);

    if (retval == FAILURE)
    { 	/* This is strange, isn't everybody allowed to leave any conference? */
	switch (kom_errno) {
	case KOM_NO_CONNECT:
	    fatal1 (CLIENT_CONNECTION_LOST, "sub_member() failed");
	    break;
	   
	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR,
		    "sub_member() failed: %d", conf_to_leave);
	}
    }

    /* Invalidate the membership-cache. */
    first_membership(0,NULL);

    RETURN (OK);
    
#undef CLEAN_UP
#undef FUNCTION
} /* do_leave_conf */



#if 0
// *  /*
// *   *  Find the position of the next conference with unread news in.
// *   *  Start looking at START_POSITION in the list of conferences
// *   *  the person is member of.
// *   */
// *  Export  Conf_no
// *  find_next_conf_with_unread (Conf_no	start_position)
// *  
// *  {
// *      static char		* this_function	= "find_next_conf_with_unread()";
// *      Conf_no		  i;		/* Loop variable */
// *      Text_list		  unread	= EMPTY_TEXT_LIST;
// *      Success		  retval;	/* Did service succeed? */
// *  
// *      
// *      i = start_position;
// *      do
// *      {
// *  	i++;
// *  	/* Wrap around to the start */
// *  	if (i >= this_person_stat.conferences.no_of_confs)
// *  	    i = 0;
// *  	zfree (unread.texts);
// *  	retval = kom_get_map (this_person_stat.conferences.confs[i].conf_no,
// *  			      this_person_stat.conferences.confs[i].last_text_read,
// *  			      MAX_LOCAL_TEXT_NO, &unread);
// *  	if (retval != OK)
// *  	{
// *  	    fatal (CLIENT_ERROR_TOO_boring,
// *  				this_function,
// *  				"kom_get_map() failed",
// *  				(fatal_info) kom_errno);
// *  	}
// *      } while (   (i != internal_state.pos_of_current_conf)
// *  	     && (unread.no_of_texts == 0));
// *  
// *      /* Clean up */
// *      zfree (unread.texts);
// *  
// *      if (i == internal_state.pos_of_current_conf)
// *  	return  0;	/* If nothing unread */
// *      else
// *  	return  i;	/* The position of the conference */
// *  }
#endif /* #if'ed out */



/*
 *  Print the name of the conference with the number CONF.
 *  Doesn't print anything if it can't find the name.
 */
Export  void
print_conf_name (Conf_no	conf)
{
    String		  name;	/* May be clobbered by longjmp */
    Kom_err		  error;

#define CLEAN_UP()				\
    { 						\
	zfree(name.string); 			\
    }

    TOPLOOP_SETJMP();

    name = conf_name (conf, &error);	/* BUG: Doesn't check for errors */
    if (s_empty (name))
	xprintf ("%s %d (%s)", txts.Number, conf, txts.Secret_c);
    else
	s_xputs (name);

    RETURNVOID;

#undef CLEAN_UP
}   /* END: print_conf_name */



/*
 *  Print a type for the conference CONF 
 */
extern  void
print_conf_type (Conference	* conf)
{
    Bool started = FALSE;

    if (conf->type.original)
    {
	xprintf ("%s ", txts.Original);
	started = TRUE;
    }

    if (conf->permitted_submitters != 0)
    {
	xprintf ("%s ", started ? txts.write_protected : txts.Write_protected);
	started = TRUE;
    }

    if (conf->type.secret)
    {
	xprintf ("%s", started ? txts.secret : txts.Secret );
    }
    else
    {
	if (conf->type.rd_prot)
	{
	    xprintf ("%s", started ? txts.closed : txts.Closed );
	}
	else
	{
	    xprintf ("%s", started ? txts.open : txts.Open );
	}
    }
}   /* END: print_conf_type() */



/*
 *  Show the status of a person.  The conference status and
 *  member-list of the person's letterbox, should already have been
 *  fetched, and be pointed to by CONF and MEMBERS respectively.  They
 *  should be released by the caller.
 */
Export  Success
show_pers_status (Conf_z_info		  conf_to_stat,
		  Conference		* conf,
		  Person		* pers		)

{

#define CLEAN_UP()     				\
    {						\
    }


    xprintf ("%s: ", txts.Last_time_in_LysKOM);
    print_time (&pers->last_login);
    xprintf (" %32T%s: %S\n", txts.As_user, pers->username);

    xprintf ("%s: %d %32T%s: %ld\n",
	     txts.Member_in_no_of_confs, pers->no_of_confs,
	     txts.Unread_letters,
	     unread_in_conf (conf_to_stat.conf_no, conf_to_stat.conf_no));

    xprintf ("%s: %lu %32T%s: %dd %dh %dm %ds\n",
	     txts.No_of_sessions, pers -> sessions,
	     txts.Total_time_present,
	     pers->total_time_present / 86400,		/* days */
	     (pers->total_time_present % 86400) / 3600,	/* hours */
	     (pers->total_time_present % 3600) / 60,	/* minutes */
	     (pers->total_time_present % 60));		/* seconds */

    xprintf ("%s: %lu %32T%s: %lu\n",
	     txts.No_of_created_texts,
	     pers->first_created_text + pers->no_of_created_texts - 1,
	     txts.No_of_created_bytes, pers->created_bytes);

    xprintf ("%s: %lu %32T%s: %lu\n",
	     txts.No_of_read_texts, pers -> read_texts,
	     txts.No_of_text_fetches, pers -> no_of_text_fetches);

    xprintf ("%s: %lu\n",
	     txts.No_of_marked_texts, pers->no_of_marks);

    xprintf ("%s: %d %32T%s: %d\n",
	     txts.No_of_created_confs, pers->created_confs,
	     txts.No_of_created_pers, pers->created_persons);

    return  OK;
}   /* END: show_pers_status() */



/*
 *  Show some short information about a conference - name, number,
 *  creator, creation time and supervisor.
 */
Export  Success
show_short_conf_status (Conf_z_info	  conf_to_stat,
			Conference	* conf		)

{
    xprintf ("%s: %s %d %32T%s: ", txts.Number,
	     conf_to_stat.type.letter_box ? txts.Person : txts.Conference,
	     conf_to_stat.conf_no, txts.Name);
    xprintf ("%S\n", conf_to_stat.name);
    xprintf ("%s: ", txts.Creation_date);
    print_time (&conf->creation_time);
    xprintf (" %32T%s: ", txts.Created_by);
    print_conf_name ((Conf_no) conf->creator);
    xprintf ("\n%s: ", txts.Supervisor);
    print_conf_name (conf->supervisor);
    newline();

    return  OK;
}   /* END: show_short_conf_status() */



/*
 *  Show some more information about a conference - number of texts,
 *  number of members, type of conf, super conf, comment conf,
 *  permitted submitters.
 */
Export  Success
show_more_conf_status (Conf_z_info	  conf_to_stat,
		       Conference	* conf		)

{
    xprintf ("%s: %5lu %32T%s: %d\n", txts.No_of_texts,
	     conf->first_local_no + conf->no_of_texts - 1,
	     txts.No_of_members, conf->no_of_members);

    xprintf ("%s: ", txts.Type);
    print_conf_type (conf);
    if (conf->super_conf != 0)
    {
	xprintf (" %32T%s: ", txts.Super_conf);
	print_conf_name (conf->super_conf);
    }
    newline();

#if 0			/* In future versions of LysKOM */
// *      if (conf->comment_conf != 0)
// *      {
// *  	xprintf ("%s: ", txts.Comment_conf);
// *  	print_conf_name (conf->comment_conf);
// *  	newline();
// *      }
#endif

    if (conf->permitted_submitters != 0)
    {
	xprintf ("%s: ", txts.Permitted_submitters);
	print_conf_name (conf->permitted_submitters);
	newline();
    }

    return  OK;
}   /* END: show_more_conf_status() */



/*
 *  Show a list of members in a conference.  The conferences status
 *  and its member-list, should already have been fetched, and be
 *  pointed to by CONF and MEMBERS respectively.  They should be
 *  released by the caller.
 */
Export  Success
show_member_list (Conf_z_info	  conf_to_stat,
		  Conference	* conf,
		  Member_list	* members)

{
    unsigned short		  count;


    xprintf ("%S %s:\n", conf_to_stat.name, txts.has_the_following_members);
    xprintf ("%s%18T%s%26T%s\n", txts.Last_time_in_conf,
	     txts.Unread, txts.Name);
    newline();

    for (count = 0; count < members->no_of_members ; count++)
    {
	Membership	  ship		= EMPTY_MEMBERSHIP;
	long		  unread;

#define CLEAN_UPPER()	release_membership (&ship)

	locate_membership (conf_to_stat.conf_no,
			   members->members[count].member,
			   &ship);
	print_time (&ship.last_time_read);
	unread = unread_in_conf (members->members[count].member,
				 conf_to_stat.conf_no);
	if (unread > 0)
	{
	    xprintf ("%18T%6ld", unread);
	}
	goto_column (26);
	print_conf_name (members->members[count].member);
	CLEAN_UPPER();
	newline();
    }

    return  OK;
}   /* END: show_member_list() */



/* Return True if MEMBERSHIP is a member of CONF_NO_LIST
 * else return False
 */
Export Bool
membership_in_conf_no_list (Membership membership,
			    Conf_no_list list)
{
    int r;

    for (r=0; r < list.no_of_confs ; r++) {
	if ( membership.conf_no == list.conf_nos[r] ) {
	    return TRUE;
	}
    }
    return FALSE;
}
