/*
 *  client/function-main.c
 *		Some very important functions for executing
 *		commands.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */


#include <config.h>

#include <stdio.h>
/* #include <time.h> included from kom-types.h */
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif


#include <kom-errno.h>
#include <kom-types.h>
#include <services.h>
#ifdef __vax__
#define __SIZE_T included from stdlib.h
#define __WCHAR_T included from stdlib.h
#endif
#include <zmalloc.h>
#include <s-string.h>
#include <s-collat.h>

#include "xoutput.h"
#include "read-line.h"
#include "commands.h"
#include "parser.h"
#include "texts.h"
#include "edit.h"
#include "cache.h"
#include "quit.h"

#include "internal.h"
#include "cmds.h"



/* This is to emphasize that a function or variable is exported. */
#define Export


/*
 *  Global variables.  They are external, but should only be used
 *  within the files 'functions-*.c'.
 */


/*  TRUE if the user wants to quit LysKOM  */
Export  Bool			  user_wants_to_quit;

/*  GOTO here when quitting */
Export jmp_buf			  toploop_jmpbuf;



/*
 *  External functions, to be exported.
 */


Export  void
command_loop (void)

{
    String		  comnd_string	= EMPTY_STRING;
    String		  arguments;
    Parse_info		  command;	/* Result from parsing */
    int			  i;		/* Index in small loops */
    volatile Conf_no	  before_command_conf_no = 0;
    volatile Local_text_no before_command_local_no = 0;

    Success (* next_command) (String argument);

    user_wants_to_quit = FALSE;

    setjmp (toploop_jmpbuf);

    newline();

    while (! user_wants_to_quit)
    {
#ifdef DEBUG
	int utl;
#endif
	handle_all_asyncs();	/* Read all asyncs. */

#define DEFAULT_COMMAND cmd_show_current_time
	next_command = DEFAULT_COMMAND; /* Default command. */

	if (next_command == DEFAULT_COMMAND) {
	    set_prompt_if_in_review(&next_command);
	}

	if (next_command == DEFAULT_COMMAND) {
	    reading_state.comment_to_read = 
		find_comment_to_read(reading_state.global_to_mark_as_read);

#ifdef DEBUG
	    xprintf("DEBUG117: reading_state.comment_to_read = %d\n",
		    reading_state.comment_to_read);
#endif

	    if (reading_state.comment_to_read == 0)
		{
		    /* If we have just read a text the texts_left count 
		       is wrong by one */
		    if ((
#ifdef DEBUG
			 utl = 
#endif
			 unread_texts_left())
			> ((before_command_conf_no == ccn
			    && before_command_local_no < reading_state.local_last_normally_read)
			   /* Detta inneb{r att det blir fel i tv} fall:
			      Om en kommentarskedja g}r ur aktuellt m|te och det bara
			      finns en ol{st kvar i m|tet s} kan vi f} prompten
			      G} till n{sta m|te - n{r vi borde f}tt prompten
			      L{sa n{sta text - eftersom m|tet vi "egentligen" {r i
			      har en text kvar.
			      Eftersom det enbart kan f|rekomma n{r vi drivit ur det
			      aktuella m|tet tycker jag f|rslusten {r ganska liten.
			      Det andra fallet {r n{r man precis l{st kommentarer och
			      man det kommer ett inl{gg till.
			      */
			   || reading_state.global_to_mark_as_read != 0 
			   ? 1 
			   : 0))
			next_command = cmd_read_next_text;
		    else {
			if (unread_confs_left() 
			    > (unread_texts_left() > 0
			       ? 1 
			       : 0))
			    next_command = cmd_goto_next_conf;
		    } 
		} else {
		    next_command = cmd_read_next_comment;
		}
#ifdef DEBUG
	    xprintf("DEBUG159: utl: %d, bccn = %d, ccn = %d, bcln = %d, rsllnr = %d\n",
		    utl, before_command_conf_no, 
		    ccn, before_command_local_no, 
		    reading_state.local_last_normally_read);
#endif
	}
	{
	    String	last_prompt = command_name( next_command );
	    newline();
	    xprintf("%S - ", last_prompt );

	    theprompt = &last_prompt;
	    read_line (&comnd_string);	/* BUG! Retval not checked! */
	    theprompt = NULL;
	}
	printed_lines = 0;
	current_column = 0;
	newline();

	before_command_conf_no = ccn;
	before_command_local_no = reading_state.local_last_normally_read;

	if (ccn != 0 && reading_state.local_last_normally_read != 0)
	    mark_as_read (ccn, 1, &reading_state.local_last_normally_read);
	if (reading_state.global_to_mark_as_read != 0) {
	    mark_global_as_read(reading_state.global_to_mark_as_read, 0);
	}
	reading_state.global_to_mark_as_read = 0;

	command = parse (comnd_string, cmd_list, TRUE, FALSE,
			 command_separators, DEFAULT_COLLAT_TAB);

	if (command.no_of_matches == 0)		/* If no matches found */
	{
	    xprintf("%s\n", txts.Not_understood_by_KOM);
	}
	else if (command.no_of_matches == 1)	/* If only one match found */
	{
	    /* Default command? */
	    if (command.indexes[0] == -1)
	    {
		(next_command)(EMPTY_STRING);
	    }
	    else
	    {
		/* Print the full name of the command */
		s_xputs(cmd_list[command.indexes[0]].name);
		xputchar (' ');
		xflush (kom_outstream);
		/* Execute the command */
		arguments = EMPTY_STRING;
		s_strcpy (&arguments, command.arguments);
		function_list [ command.indexes [0] ] (arguments);
	    }
	}
	else		/* More than one match found */
	{
	    xprintf("%s. %s:\n",
		   txts.Command_was_ambigous,
		   txts.You_might_mean_one_of_these);
	    for ( i = 0 ;  i < command.no_of_matches ;  i++ )
	    {
		s_xputs (cmd_list[command.indexes[i]].name);
		newline();
	    }
	}
	
	/* Release memory allocated for index table */
	zfree (command.indexes);
    }
    s_clear (&comnd_string);
}   /* END: command_loop() */



/*
 *  The parameter 'ARGUMENT' in all the external functions, is a
 *  String containing the rest of the line, following the command.
 *  E g if the user gave the command "Skapa m|te Inl{gg }t mig",
 *  then the String "Inl{gg }t mig" is passed as the parameter
 *  'ARGUMENT' to the function responsible for the command 'skapa
 *  m|te'.  Leading and trailing spaces are skipped.  EMPTY_STRING
 *  is passed if no arguments to the command was given.
 */


/*  Log out from server and exit client.  */
Export  Success
cmd_quit_LysKOM (String   argument)
{
#define CLEAN_UP()	do { s_clear (&argument); } while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    kom_logout();
    user_wants_to_quit = TRUE;
    xprintf ("\n%s. %s %s.\n",
	    txts.Thanks_for_this_time,
	    "------------------------------",
	    txts.Welcome_back);

    RETURN (OK);
#undef CLEAN_UP
}




Export  Success
cmd_not_yet_implemented (String   argument)
{
    xprintf("%s.\n", txts.Not_implemented_yet);
    s_clear (&argument);
    return  OK;
}
