/*
 *  client/functions-read.c
 *		The functions to actually execute the for the
 *		client.  Commands to read texts.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */

#include <config.h>

#include <stdio.h>
/* #include <time.h> is included from kom-types.h */
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#include <kom-errno.h>
#include <kom-types.h>
#include <services.h>
#ifdef __vax__
#define __SIZE_T included from stddef.h
#define __WCHAR_T included from stddef.h
#endif
#include <zmalloc.h>
#include <s-string.h>
#include <s-collat.h>

#include "conf.h"

#include "xoutput.h"
#include "copy.h"
#include "read-line.h"
#include "commands.h"
#include "parser.h"
#include "texts.h"
#include "edit.h"
#include "misc-parser.h"
#include "error.h"
#include "cache.h"
#include "quit.h"

#include "internal.h"
#include "cmds.h"


/* This is to emphasize that an object is exported. */
#define Export

/* We are keeping track of whats going on here. */
Export Internal_state_t	reading_state;

/* Local function. */
static int recursively_mark_texts_as_read(Text_no);


/* Prototypes */
void push_read_comments_to(Text_no);
void pop_to_do(void);
To_do_stack peek_to_do(void);
int unread_confs_left(void);
int unread_texts_left(void);
int decrease_unread_texts_left(void);

/*
 *  The parameter 'ARGUMENT' in all the external functions, is a
 *  String containing the rest of the line, following the command.
 *  E g if the user gave the command "Skapa m|te Inl{gg }t mig",
 *  then the String "Inl{gg }t mig" is passed as the parameter
 *  'ARGUMENT' to the function responsible for the command 'skapa
 *  m|te'.  Leading and trailing spaces are skipped.  EMPTY_STRING
 *  is passed if no arguments to the command was given.
 */


/*  Find and display text(s) according to strange search patterns.  */
Export  Success
cmd_find_and_view_text (String   argument)
{
    Text_no		  wanted_text;
    const String	  the_commented	= s_fcrea_str(txts.the_commented);
    String_size		  error_check;	/* First uncoverted char str->num */
    Success		  retval;


#define CLEAN_UP()	do {	s_clear (&argument);	} while (0)

    TOPLOOP_SETJMP();

    newline();

    if (s_empty(argument))
    {
	xprintf("%s?\n= ", txts.What_texts_to_view);
	read_line (&argument);		/* BUG! Check retval! */
	newline();
    }
    argument = s_strip_trailing (argument, command_separators);

    /* A number, and nothing else? */
    wanted_text = (Text_no) s_strtol (argument, &error_check,
				      DEFAULT_NUMBER_BASE);
    if (error_check < s_strlen (argument))
    {
	/* No, it wasn't a number.  Then try "(the) commented" */
	/* ??? Check this out */
	if (! s_usr_strhead (argument, the_commented, DEFAULT_COLLAT_TAB))
	{
	    xprintf ("%s: ", txts.Unknown_argument);
	    s_puts(argument);
	    newline();
	    RETURN (FAILURE);
	}
	cmd_not_yet_implemented (argument);
	RETURN (FAILURE);
    }
    else
    {
	/* OK, it was a number */
	retval = show_text_no (wanted_text);
	RETURN (retval);
    }

    /*NOTREACHED*/
    RETURN (OK);
#undef CLEAN_UP
}




/*  Read the next text in straight order  */
Export  Success
cmd_read_next_text (String   argument)

{
#define FUNCTION "read_next_text()"

    Success		  exit_value;
    Local_text_no	  text_to_read;
    Text_no		  global_no_to_read;
    Kom_err		  error;


#define CLEAN_UP()	do {	} while (0)


    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    if (find_next_unread_text (&text_to_read) == FAILURE)
    {
	switch (kom_errno)
	{
	case  KOM_UNDEF_CONF:
	case  KOM_PERM:
	    xprintf ("%s.\n", txts.Not_allowed_to_read_anymore);
	    break;

	default:
	    fatal2 (CLIENT_SHOULDNT_HAPPEN,
		    "find_next_unread_text(%d) returns bogus",
		    text_to_read);
	    break;
	}

	RETURN (FAILURE);
    }

    global_no_to_read = map_local_to_global (ccn, text_to_read, &error);
    /* BUG: Doesn't check for error */
    exit_value = show_text_no (global_no_to_read);

    /* The text will be marked as read when the next command is given. */
    /* But the counting is done here. */
    decrease_unread_texts_left();

    if (exit_value == OK) {
	reading_state.last_normally_read = global_no_to_read;
	reading_state.local_last_normally_read = text_to_read;
	push_read_comments_to(global_no_to_read);
    }

    RETURN (exit_value);

#undef CLEAN_UP
#undef FUNCTION
}


/*  Read the commented text  */
extern Success
cmd_read_commented_text (String  argument)
{
#define FUNCTION "read_commented_text()"

    Text_no		  text_to_search_from;
    volatile Text_no	  text_to_read = 0;
    Text_stat		  text_stat = EMPTY_TEXT_STAT;

#define CLEAN_UP()	do { s_clear (&argument); \
			     release_text_stat(&text_stat); } while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    text_to_search_from = reading_state.last_viewed;

    if ( text_to_search_from
	 && kom_get_text_stat (text_to_search_from,
			       &text_stat ) != FAILURE)
    {
	Misc_info     * misc;
	unsigned short	r;

	newline();
	for (misc = text_stat.misc_items , r = 0;
	     r < text_stat.no_of_misc && !text_to_read;
	     r++, misc++)
	{
	    if (misc->type == comm_to)
	    {
		text_to_read = misc->datum.comment_to;
	    }
	}

	if (text_to_read)
	{
	    RETURN (show_text_no (text_to_read));
	}
	else 
	{
	    xprintf("%s.\n", txts.Text_doesnt_exist);
	    RETURN (OK);
	}
    }
    
    RETURN (FAILURE);
#undef CLEAN_UP
#undef FUNCTION
}


/* Read the next text in comment tree.
 *
 * This is easy since we have already calculated what text to read.
 */
Export  Success
cmd_read_next_comment (String   argument)

{
#define FUNCTION "cmd_read_next_comment()"

    Success		  exit_value;
    Text_no		  global_no_to_read;


#define CLEAN_UP()	do {	} while (0)


    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    if (reading_state.comment_to_read == 0) 
    {
	RETURN(FAILURE);
    }

    global_no_to_read = reading_state.comment_to_read;
    /* BUG: Doesn't check for error */
    exit_value = show_text_no (global_no_to_read);

    /* The text will be marked as read when the next command is given. */
    /* But the counting is done here. */
    decrease_unread_texts_left();

    if (exit_value == OK) {
	reading_state.last_normally_read = global_no_to_read;
	reading_state.local_last_normally_read = 0;
	reading_state.global_to_mark_as_read = global_no_to_read;
	push_read_comments_to(global_no_to_read);
    }

    RETURN (exit_value);

#undef CLEAN_UP
#undef FUNCTION
}


/* Returns the next unread comment to read.
   0 if none left.
 */
Text_no
find_comment_to_read(Text_no skipthis)
{
#define FUNCTION "find_comment_to_read()"
    Text_stat * volatile  stat = NULL;
    Kom_err		  error;
    Text_no		  commentno;

#define CLEAN_UP()	do { release_text_stat(stat); \
			     if (stat != NULL) zfree(stat); } while (0)

    TOPLOOP_SETJMP();

    while (peek_to_do()) 
    {
	const Misc_info 	* misc;
	Misc_info_group		  misc_data;

	/* Free before. */
	if (stat != NULL)
	{
	    release_text_stat(stat);
	    zfree(stat);
	}

	/* We have something to do. */
	stat = text_stat(peek_to_do()->info.comments_to.textnumber, &error);
	
	if (stat == NULL)
	{
	    switch(error) {
	    default:
		/* We have som problem here. */
		CLEAN_UP();
		pop_to_do();
		continue;
	    }
	    /*NOTREACHED*/
	}

	/* Lets doit for footnotes. */
	for (misc = stat->misc_items;
	     (misc_data = parse_next_misc(&misc,
					  stat->misc_items + 
					  stat->no_of_misc)).type
	     != m_end_of_list;
	     )
	{
#ifdef HAVE_ASSERT_H
	    assert(misc_data.type != m_error);
#endif

	    if (misc_data.type == m_footn_in)
	    {
		commentno = misc_data.footnoted_in;
		if (commentno == skipthis)
		    continue;
		/* Time to check if this article is unread by the user */
		if (has_been_read(commentno) == FALSE)
		{
		    RETURN(commentno);
		}
	    }
	}
	/* Lets do it for comments */
	for (misc = stat->misc_items;
	     (misc_data = parse_next_misc(&misc,
					  stat->misc_items + 
					  stat->no_of_misc)).type
	     != m_end_of_list;
	     )
	{
	    if (misc_data.type == m_comm_in)
	    {
		commentno = misc_data.commented_in;
		if (commentno == skipthis)
		    continue;
		/* Time to check if this article is unread by the user */
		if (has_been_read(commentno) == FALSE)
		{
		    RETURN(commentno);
		}
	    }
	}

	pop_to_do();
    }
    RETURN(0);
#undef FUNCTION
#undef CLEAN_UP
}


/*  Jump comments to the last read article
 *  The search is done recursively.
 *  Loops are handled in the way that we only continue the recursive
 *  search if the comment is newer.
 */
extern Success
cmd_jump_tree (String	argument)
{
#define FUNCTION "cmd_jump_tree()"
    
#define CLEAN_UP()	do { ; } while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    newline();
    if (reading_state.last_viewed) {
	int removed;

	removed = recursively_mark_texts_as_read(reading_state.last_viewed);
	removed--;
	if (removed == 0)
	    xprintf(txts.no_jumped);
	else if (removed == 1)
	    xprintf(txts.one_jumped);
	else if (removed > 1)
	    xprintf(txts.several_jumped, removed);

	RETURN (OK);
    }

    RETURN (FAILURE);
#undef CLEAN_UP
#undef FUNCTION
}


/*  Goto the next conference with unread news. */
extern Success
cmd_goto_next_conf (String   argument)
{
#define FUNCTION "cmd_goto_next_conf()"

    Kom_err		  error;

    Conf_no_list	  list = EMPTY_CONF_NO_LIST;
    Membership 		  ship = EMPTY_MEMBERSHIP;
    Membership_list	  ship_list = EMPTY_MEMBERSHIP_LIST;
    int 		  pos;
    
#define CLEAN_UP()	do {	release_conf_no_list (&list); \
				release_membership_list (&ship_list); \
			} while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    review_reset();

    if (kom_get_unread_confs (cpn, &list) == FAILURE)
    {
	switch (kom_errno)
	{
	case  KOM_OUT_OF_MEMORY:
	    fatal1 (CLIENT_OUT_OF_MEMORY, "get_conf_no_list() failed");
	    break;

	case  KOM_NO_CONNECT:
	    fatal1 (CLIENT_SERVER_ERROR, "get_conf_no_list() failed");
	    break;

	default:
	    fatal1 (CLIENT_UNEXPECTED_SERVER_ERROR,
		    "get_conf_no_list() failed");
	    break;
	}
    }

    for (pos = first_membership(cpn, &ship);
	 pos != -1;
	 pos = next_membership(&ship))
    {
	if ( membership_in_conf_no_list(ship, list) )
	{
	    int unread;

	    unread = conf_last(ship.conf_no, &error)
		- (ship.last_text_read + ship.no_of_read);
	    if (unread > 0) {
		print_conf_name(ship.conf_no);
		xprintf(" - %d %s.\n", unread, 
			(unread == 1 ? txts.unread_singular
			             : txts.unread_plural));
		goto_conference(ship.conf_no);
		RETURN (OK);
	    } else {
		/* There are some problem with the cache, probably
		   a text added to a conf that makes kom_get_unread_confs 
		   return something while the cache doesn't contain anything.
		   Lets refetch that conference. */
		/* BUG * this should be fixed with a smarter cache */
		Conference * conf;

		conf = conf_stat (ship.conf_no, &error);
		zfree (release_conf (conf));
	    }
	}
    }
    
    RETURN (FAILURE);

#undef CLEAN_UP
#undef FUNCTION
}


/*  Endast, mark text as read quickly!
 * 
 *  If the argument given is smaller than the amount of unread, then we
 *  calculate the amount of articles that are to be marked as read and
 *  move up the last_read-counter with the special request.
 *
 *  If the argument given is bigger than the amount of unread, the we
 *  move down the last_read-counter
 */
Export  Success
cmd_only (String   argument)
{
#define FUNCTION "cmd_only()"

    int			  unread_wanted;
    String_size           error_check;  /* First uncoverted char str->num */
    Success		  retval;


#define CLEAN_UP()	do { s_clear(&argument); } while (0)

    TOPLOOP_SETJMP();

    newline();

    if (ccn == 0)
    {
	xprintf(txts.Not_in_any_conf);
	RETURN (FAILURE);
    }

    if (s_empty(argument))
    {
        xprintf (txts.How_many_to_view);
        read_line (&argument);          /* BUG! Check retval! */
	newline();
    }
    argument = s_strip_trailing (argument, command_separators);

    /* A number, and nothing else? */
    unread_wanted = s_strtol (argument, &error_check,
			      DEFAULT_NUMBER_BASE);
    if (error_check < s_strlen (argument))
    {
        xprintf ("%s: ", txts.Unknown_argument);
        s_puts(argument);
        newline();
        RETURN (FAILURE);
    }
    else
    {
	/* Forget all cached information about that conf. */
	Conf_no saved_ccn = ccn;

	cache_change_person(cpn); /* Reset all the cache. */
	(void) unread_texts_left(); /* Reset the count. */
	ccn = saved_ccn;
	
	/* Just set the new value */
	retval = kom_set_unread(ccn, unread_wanted);

	if (retval == FAILURE)
	{
	    switch (kom_errno)
	    {
	    case  KOM_UNDEF_CONF:
	    case  KOM_PERM:
		xprintf ("%s.\n", txts.Not_allowed_to_read_anymore);
		break;

	    default:
		fatal3 (CLIENT_SHOULDNT_HAPPEN,
			"kom_set_unread(%d, %d) returns bogus",
			ccn, unread_wanted);
		break;
	    }

	    RETURN (FAILURE);
	}

    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}

/* Display all new articles. */
Export  Success
cmd_all (String   argument)
{
#define FUNCTION "cmd_all()"

    volatile Conf_no		before_command_conf_no = 0;
    volatile Local_text_no	before_command_local_no = 0;

    Success (* next_command) (String argument);

#define CLEAN_UP()	do { s_clear(&argument); } while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    newline();

    for(;;) {
#ifdef DEBUG
	int utl;
#endif
	handle_all_asyncs();	/* Read all asyncs. */

	next_command = cmd_show_current_time; /* Default command. */

	reading_state.comment_to_read = 
	    find_comment_to_read(reading_state.global_to_mark_as_read);

#ifdef DEBUG
	xprintf("DEBUG117: reading_state.comment_to_read = %d\n",
		reading_state.comment_to_read);
#endif

	if (reading_state.comment_to_read == 0)
	{
	    /* If we have just read a text the texts_left count 
	       is wrong by one */
	    if ((
#ifdef DEBUG
		 utl = 
#endif
		 unread_texts_left())
		> ((before_command_conf_no == ccn
		    && before_command_local_no < reading_state.local_last_normally_read)
		   /* Detta inneb{r att det blir fel i tv} fall:
		      Om en kommentarskedja g}r ur aktuellt m|te och det bara
		      finns en ol{st kvar i m|tet s} kan vi f} prompten
		      G} till n{sta m|te - n{r vi borde f}tt prompten
		      L{sa n{sta text - eftersom m|tet vi "egentligen" {r i
		      har en text kvar.
		      Eftersom det enbart kan f|rekomma n{r vi drivit ur det
		      aktuella m|tet tycker jag f|rslusten {r ganska liten.
		      Det andra fallet {r n{r man precis l{st kommentarer och
		      man det kommer ett inl{gg till.
		    */
		   || reading_state.global_to_mark_as_read != 0 
		   ? 1 
		   : 0))
		next_command = cmd_read_next_text;
	    else {
		if (unread_confs_left() 
		    > (unread_texts_left() > 0
		       ? 1 
		       : 0))
		    next_command = cmd_goto_next_conf;
	    } 
	} else {
	    next_command = cmd_read_next_comment;
	}
#ifdef DEBUG
	xprintf("DEBUG159: utl: %d, bccn = %d, ccn = %d, bcln = %d, rsllnr = %d\n",
		utl, before_command_conf_no, 
		ccn, before_command_local_no, 
		reading_state.local_last_normally_read);
#endif
	printed_lines = 0;
	current_column = 0;
	newline();

	before_command_conf_no = ccn;
	before_command_local_no = reading_state.local_last_normally_read;

	if (ccn != 0 && reading_state.local_last_normally_read != 0)
	    mark_as_read (ccn, 1, &reading_state.local_last_normally_read);
	if (reading_state.global_to_mark_as_read != 0) {
	    mark_global_as_read(reading_state.global_to_mark_as_read, 0);
	}
	reading_state.global_to_mark_as_read = 0;

        if (next_command == cmd_show_current_time) break;

        (next_command)(EMPTY_STRING);
    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}   /* END: command_loop() */

/* ==================================================================== */
/*			Internal functions				*/


/*
 *  Fetch and show the text WANTED if the server is willing to let
 *  the person see it.  Currently only one format for showing the
 *  texts - no options available.
 */
Export  Success
show_text_no (Text_no	wanted)

{
/*    static char		* this_function	= "show_text_no()";*/

    Text_stat	*volatile stat		= NULL;
    String		  text		= EMPTY_STRING;
    Kom_err		  error;


#define CLEAN_UP()	do {				\
	if (stat != NULL)				\
	    zfree (stat->misc_items);			\
	zfree (stat);					\
	s_clear (&text);					} while (0)

    TOPLOOP_SETJMP();

    if (  (stat = text_stat (wanted, &error)) == NULL
	|| ( s_empty (text = get_text (wanted, 0, END_OF_STRING, &error))
	    && (error != KOM_NO_ERROR) ))
    {
	xprintf ("%s.\n", txts.Text_doesnt_exist);
	RETURN (FAILURE);
    }

    newline();
    print_text_header (wanted, stat);
    print_misc_list (stat->misc_items, stat->no_of_misc,
		     PRINT_RECPT | PRINT_CC_RECPT);
    print_misc_list (stat->misc_items, stat->no_of_misc,
		     PRINT_COMM_TO | PRINT_FOOTN_TO);
    print_subject_line (txts.Subject, text);
    xprintf ("------------------------------------------------------------\n");
    print_rest_of_text (text);
    xprintf ("(%lu)------------------------------\n", wanted);
    print_misc_list (stat->misc_items, stat->no_of_misc,
		     PRINT_COMM_IN | PRINT_FOOTN_IN);
    reading_state.last_viewed = wanted;

    RETURN (OK);

#undef CLEAN_UP
}



Export  void
print_text_header (Text_no		  number,
		   const Text_stat	* stat   )
{
    xprintf ("(%lu) ", number);
    print_time (&stat->creation_time);
    xprintf (" /%d %s/ ", stat->no_of_lines, txts.lines);
    print_conf_name ((Conf_no) stat->author);
    newline();
}



Export  void
print_misc_list (const Misc_info	* misc_list,
		 unsigned short		  no_of_misc,
		 unsigned short		  what_to_print)

{
#define FUNCTION "print_misc_list()"

    const Misc_info	* stop_ptr;	/* Address of element beyond last */
    Misc_info_group	  misc_data;
    Text_stat		* stat2;	/* For info about other texts */
    Kom_err		  error;


    stop_ptr = misc_list + no_of_misc;
    while (  (misc_data = parse_next_misc (&misc_list, stop_ptr)).type
	   != m_end_of_list)
    {
	switch (misc_data.type)
	{
	case  m_recpt:
	    if (! (what_to_print & PRINT_RECPT))
		break;
	    /* Else fall through */

	case  m_cc_recpt:
	    if (! (what_to_print & (PRINT_RECPT | PRINT_CC_RECPT)))
		break;

	    if (misc_data.type == m_recpt) {
		if (is_sent_by (misc_data))
		    xprintf ("%s: ", txts.Carbon_copy);
		else
		    xprintf ("%s: ", txts.Recipient);
		print_conf_name (misc_data.recipient);
	    }
	    else {
		xprintf ("%s: ", txts.Extra_copy);
		print_conf_name (misc_data.cc_recipient);
	    }
	    if (misc_data.local_no) /* If we have a local_no */
		xprintf (" <%lu>", misc_data.local_no);

	    if (is_received (misc_data))
	    {
		xprintf (" -- %s: ", txts.Received);
		print_time (&misc_data.received_at);
	    }

	    newline();

	    if (is_sent_by (misc_data))
	    {
		xprintf (" %s: ", txts.Sender);
		print_conf_name ((Conf_no) misc_data.sender);
	    }
	    if (is_sent (misc_data))
	    {
		xprintf (" -- %s: ", txts.Sent_at);
		print_time (&misc_data.sent_at);
		newline();
	    }

	    break;


	case  m_comm_to:
	    if (! (what_to_print & PRINT_COMM_TO))
		break;

	    xprintf ("%s %lu", txts.Comment_to_text,
		     misc_data.comment_to);
	    if ((stat2 = text_stat (misc_data.comment_to, &error)) != NULL)
	    {
		xprintf (" %s ", txts.by);
		print_conf_name ((Conf_no) stat2->author);
		zfree (stat2->misc_items);
		zfree (stat2);
	    }
	    newline();
	    break;


	case  m_footn_to:
	    if (! (what_to_print & PRINT_FOOTN_TO))
		break;

	    xprintf ("%s %lu", txts.Footnote_to_text,
		     misc_data.footnote_to);
	    if ((stat2 = text_stat (misc_data.footnote_to, &error)) != NULL)
	    {
		xprintf (" %s ", txts.by);
		print_conf_name ((Conf_no) stat2->author);
		zfree (stat2->misc_items);
		zfree (stat2);
	    }
	    newline();
	    break;


	case  m_comm_in:
	    if (! (what_to_print & PRINT_COMM_IN))
		break;

	    xprintf ("(%s %lu", txts.Comment_in_text,
		     misc_data.commented_in);
	    if ((stat2 = text_stat (misc_data.commented_in, &error)) != NULL)
	    {
		xprintf (" %s ", txts.by);
		print_conf_name ((Conf_no) stat2->author);
		zfree (stat2->misc_items);
		zfree (stat2);
	    }
	    xputchar(')');
	    newline();
	    break;


	case  m_footn_in:
	    if (! (what_to_print & PRINT_FOOTN_IN))
		break;

	    xprintf ("(%s %lu", txts.Footnote_in_text,
		     misc_data.footnoted_in);
	    if ((stat2 = text_stat (misc_data.footnoted_in, &error)) != NULL)
	    {
		xprintf (" %s ", txts.by);
		print_conf_name ((Conf_no) stat2->author);
		zfree (stat2->misc_items);
		zfree (stat2);
	    }
	    xputchar(')');
	    newline();
	    break;


	case  m_error:
	    server_bug ("Bad misc-items", FUNCTION,
			(bug_info) misc_list);
	    break;


	case  m_end_of_list:
	    abort();	/* This case is taken care of by the while */
	}
    }	/* while (! end-of-misc-list) */
#undef FUNCTION
}



Export  void
print_subject_line (const unsigned char	* subject_text,
		    const String	text		)
{
    String_size		  end_of_line_pos;


    xprintf ("%s:  ", subject_text);
    end_of_line_pos = s_strchr (text, '\n', 0);
    s_xputs (s_fsubstr (text, 0, (end_of_line_pos == -1
				  ? END_OF_STRING
				  : end_of_line_pos)));
}



Export  void
print_rest_of_text (const String	text)
{
    String_size		  end_of_line_pos;


    end_of_line_pos = s_strchr (text, '\n', 0);
    s_xputs (s_fsubstr (text, end_of_line_pos + 1, END_OF_STRING));
    if (text.len == 0 || text.string[s_strlen (text) - 1] != '\n')
	newline();
}



Export  void
print_time (const struct tm	* time)
{
    xprintf ("%04d-%02d-%02d %02d.%02d",
	     time->tm_year+1900, time->tm_mon+1,
	     time->tm_mday, time->tm_hour, time->tm_min);
}



/*
 * Recursively mark all comments to the text TNO read.
 * Returns how many we have marked.
 */
static int
recursively_mark_texts_as_read(Text_no tno)
{
    Text_stat		text_stat = EMPTY_TEXT_STAT;
    volatile int	jumped = 0;

#define CLEAN_UP()	do { release_text_stat(&text_stat); } while (0)

    TOPLOOP_SETJMP();

    if (tno			/* Check for stupid calls. */
	&& kom_get_text_stat (tno, &text_stat) != FAILURE)
    {
	const Misc_info       * misc;
	Misc_info_group 	misc_data;

	/* Do the recursive search */
	for (misc = text_stat.misc_items;
	     (misc_data = parse_next_misc(&misc,
					  text_stat.misc_items +
					  text_stat.no_of_misc)).type
	     != m_end_of_list;
	     )
	{
	    if (misc_data.type == m_comm_in)
	    {
		if (misc_data.commented_in > tno)
		    jumped += recursively_mark_texts_as_read(misc_data.commented_in);
	    }
	}
	
	/* Mark this article as read. */
	
	(void) decrease_unread_texts_left(); /* This is a simple 
						approximation. We suppose we 
						are in the same conf. */
	mark_global_as_read(tno, 0); /* That was easy. */
	jumped ++;
    }
    RETURN(jumped);
#undef CLEAN_UP
}


/*
 *  Find the next unread text in the current conference and
 *  put the result in NEXT_TO_READ.  
 *  Beware, this could give an non-existing text.
 *
 *  Returns OK if everything went fine, and FAILURE if something
 *  (non-fatal) went wrong, e g couldn't get the status of the
 *  current conference.  The global variable KOM_ERRNO is set
 *  to the error.
 */
/* BUG: does not check if a text already has been read,
   via a "]terse"-command, or in another conference	*/
Export  Success
find_next_unread_text (Local_text_no	* next_to_read)

{
#define FUNCTION "find_next_unread_text()"
    Kom_err		error;
    Local_text_no	to_read[1];
    Text_no 		text;
    Membership		ship;

#define CLEAN_UP()	do {	} while (0)

    TOPLOOP_SETJMP();

    if (locate_membership(ccn, cpn, &ship) == -1)
    {			/* There is a problem with this one... */
	fatal2 (CLIENT_ERROR_TOO_boring,
		"membership of current conf (%s) not found. Conf deleted, you are thrown out...",
		ccn);
    }

    to_read[0] = ship.last_text_read + 1;

    while (to_read[0] <= conf_last(ccn, &error)
	   && (text = map_local_to_global (ccn,to_read[0],&error)) == 0
	   && kom_errno == KOM_NO_ERROR)
    {
	mark_as_read( ccn, 1, to_read ); /* This will modify ship. */
	decrease_unread_texts_left();
	
	/* Reload the ship. */
	if (locate_membership(ccn, cpn, &ship) == -1)
	{			/* There is a problem with this one... */
	    fatal2 (CLIENT_ERROR_TOO_boring,
		    "membership of current conf (%s) not found. Conf deleted, you are thrown out...",
		    ccn);
	}

	to_read[0] = ship.last_text_read + 1;
    }
    * next_to_read = to_read[0];

    RETURN (OK);

#undef CLEAN_UP
#undef FUNCTION
}


/*
 * Check if a text has been read by the user.
 * Returns False if unread.
 */
Export  Bool
has_been_read (Text_no		  text)
{
#define FUNCTION "has_been_read()"
    Text_stat * volatile  stat = NULL;
    Kom_err		  error;
    const Misc_info 	* misc;
    Misc_info_group	  misc_data;
    Membership		  membship = EMPTY_MEMBERSHIP;

#define CLEAN_UP()	do { release_text_stat(stat); \
			     if (stat != NULL) zfree(stat); \
			 } while (0)
/* Dont release the membership here. It is not marked. */

    TOPLOOP_SETJMP();

    stat = text_stat(text, &error);

    if (stat == NULL) {
	switch(error) {
	default:
	    /* Something went wrong. Consider the text as read. */
	    RETURN(TRUE);
	}
	/*NOTREACHED*/
    }

    for (misc = stat->misc_items;
	 (misc_data = parse_next_misc(&misc,
				      stat->misc_items +
				      stat->no_of_misc)).type
	 != m_end_of_list;
	 )
    /* This for-loop is really a joke because we RETURN out of it at the first
       conference. */
    {
	if (misc_data.type == m_recpt
	    || misc_data.type == m_cc_recpt)
	{
	    if (locate_membership ((misc_data.type == m_recpt
				    ? misc_data.recipient
				    : misc_data.cc_recipient),
				   cpn, &membship) >= 0)
	    { /* At least we are members. */
		int r;

		if (misc_data.local_no <= membship.last_text_read)
		{
		    RETURN(TRUE); /* Pheuw */
		}
		if (membship.no_of_read == 0)
		{
		    RETURN(FALSE); /* If it hasn't been read here, 
				      it hasn't been read anywhere. */
		}
		for (r = 0; r < membship.no_of_read; r++) {
		    if (membship.read_texts[r] == misc_data.local_no)
		    {
			RETURN(TRUE);
		    }
		}
		RETURN(FALSE);	/* It hasn't been read. */
	    }
	}
    }
    RETURN (TRUE);		/* No recipient. This is read. */
#undef FUNCTION
#undef CLEAN_UP
}



/*
 *  Report the number of unread texts the person WHO has in
 *  conference WHERE.
 */
Export  long
unread_in_conf (Pers_no	    who,
		Conf_no	    where)

{
    Membership		  membship	= EMPTY_MEMBERSHIP;
    long		  no_of_unread;
    Kom_err		  error;

#define CLEAN_UP()	do { } while (0)

    TOPLOOP_SETJMP();

    if (locate_membership (where, who, &membship) < 0)
    {
	RETURN (-1);
    }

    no_of_unread = (conf_last(where, &error)
		    - membship.last_text_read - membship.no_of_read);

    RETURN (no_of_unread);

#undef CLEAN_UP
}   /* END: unread_in_conf() */




#define NOTYETREAD -1

/* How many articles are there left to read in this conference? */
static int more_to_read = NOTYETREAD;
static Conf_no more_to_read_for_conf; 

/* Returns the value of more_confs i.e the number of unread confs left.
 */
int
unread_confs_left()
{
#define FUNCTION "unread_confs_left()"

    Conf_no_list	  list = EMPTY_CONF_NO_LIST;
    int			  count_of_uconfs;

    if (kom_get_unread_confs (cpn, &list) == FAILURE) {
	switch (kom_errno) {
	case  KOM_OUT_OF_MEMORY:
	    fatal1 (CLIENT_OUT_OF_MEMORY, "get_conf_no_list() failed");
	    break;

	case  KOM_NO_CONNECT:
	    fatal1 (CLIENT_SERVER_ERROR, "get_conf_no_list() failed");
	    break;

	default:
	    fatal1 (CLIENT_UNEXPECTED_SERVER_ERROR, 
		    "get_conf_no_list() failed");
	    break;
	}
    }

    more_to_read = NOTYETREAD;
    more_to_read_for_conf = 0;
    count_of_uconfs = list.no_of_confs;
    release_conf_no_list(&list); 
    return count_of_uconfs;
#undef FUNCTION
}



/* Returns the amount of texts left in this conf to be read.
 * If we havnt read the amount of unread then get them.
 * Returns 0 if not in any conf or there are 0 unread left.
 */
int
unread_texts_left()
{
    if (ccn == 0) {
	more_to_read = NOTYETREAD;
	more_to_read_for_conf = 0;
	return 0;
    }
    if (more_to_read == NOTYETREAD
	|| ccn != more_to_read_for_conf
	|| more_to_read == 0
	|| more_to_read == 1 /* We need to know if it really is one or
				if the counter just happened to be 1 */) {
	more_to_read = unread_in_conf(cpn,ccn);
	more_to_read_for_conf = ccn;
    }
    return more_to_read;
}


/* Decreases the amount of unread texts.
 * This is to get a quicker count.
 */
int
decrease_unread_texts_left()
{
    if (more_to_read > 0)
	--more_to_read;
    return more_to_read;
}




/* Here are the functions maintaining the to_do_stack:
   (void) push_read_comments_to(Text_no);
   (void) pop_to_do();
   (To_do_stack) peek_to_do();
 */
void
push_read_comments_to(Text_no	tno)
{
    To_do_stack	new_elem = (To_do_stack)malloc(sizeof(struct to_do_stack));

    new_elem->doing = reading_comments_to_this_text;
    new_elem->info.comments_to.textnumber = tno;
    new_elem->next = reading_state.thingstodo;

    reading_state.thingstodo = new_elem;
}

void
pop_to_do()
{
    if (reading_state.thingstodo)
    {
	To_do_stack old_elem;
	
	old_elem = reading_state.thingstodo;
	reading_state.thingstodo = old_elem->next;
	free(old_elem);
    }
}
	

To_do_stack
peek_to_do()
{
    return reading_state.thingstodo;
}
