/*
 *  functions-edit.c
 *		Functions for the commands that edit a text being written.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */

#include <config.h>

#include <stdio.h>
/* #include <time.h> included from kom-types.h */
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef __vax__
#define __SIZE_T included from string.h
#endif
#include <kom-types.h>	/* Before zmalloc.h */
#include <zmalloc.h>
#include <s-string.h>
#include <s-collat.h>

#include <kom-errno.h>
#include <services.h>

#include "conf.h"

/* Must be included before "read-line.h" on a vax because it includes <sgtty.h>
 * and that does not work after inclusion of <sys/ioctl.h> because of
 * stupid include-file construction.
 */
#include "xoutput.h"

#include "read-line.h"
#include "commands.h"
#include "parser.h"
#include "texts.h"
#include "edit.h"
#include "error.h"
#include "cache.h"
#include "quit.h"

#include "internal.h"
#include "cmds.h"


/* This is to emphasize that a function is exported. */
#define Export


/*  ==================================================================  */
/*			    Exported functions				*/


Export  Success
cmd_mark_permanently (String	  argument)

{
    Success			  retval;
    Text_no			  text_to_mark;
    String_size			  error_check;

#define CLEAN_UP()	do	{	s_clear (&argument);	}while (0)

    TOPLOOP_SETJMP();

    if (s_empty (argument))
    {
	text_to_mark = reading_state.last_viewed;
    }
    else
    {
	text_to_mark = s_strtol (argument, &error_check,
				 DEFAULT_NUMBER_BASE);
	if (error_check < s_strlen (argument))
	{
	    /*  No, it wasn't a number.  */
	    xprintf ("%s:  ", txts.Unknown_argument);
	    s_puts (argument);
	    newline();
	    RETURN (FAILURE);
	}
    }
    xprintf ("%d", text_to_mark);
    newline();

    retval = kom_mark_text (text_to_mark, 255);		/* 255 is a BUG! */
    if (retval != OK)
    {
	xprintf ("%s.\n", txts.Cant_mark_text);
	RETURN (FAILURE);
    }

    RETURN (OK);

#undef CLEAN_UP
}   /* END: cmd_mark_permanently() */



Export	Success
cmd_delete_text (String		  argument)

{
    Text_no			  text_to_delete;
    String_size			  error_check;

#define CLEAN_UP()	do	{ 	s_clear (&argument);	} while (0)

    TOPLOOP_SETJMP();

    if (s_empty (argument))
    {	
	xprintf(txts.Give_the_textno_to_delete);
	RETURN (FAILURE);
    }
    else
    {
	text_to_delete = s_strtol (argument, &error_check,
				   DEFAULT_NUMBER_BASE);
	if (error_check < s_strlen (argument))
	{
	    /*  No, it wasn't a number.  */
	    xprintf ("%s:  ", txts.Unknown_argument);
	    s_puts (argument);
	    newline();
	    RETURN (FAILURE);
	}
	xprintf ("%d", text_to_delete);
    }

    newline();
    xprintf (txts.Really_sure_delete_text, text_to_delete);

    if (yes_or_no_p())
    {
	if (kom_delete_text( text_to_delete ) != FAILURE)
	{
	    RETURN (OK);
	}
	else
	{
	    xprintf(txts.Delete_text_failed, text_to_delete);
	    RETURN (FAILURE);
	}
    }
    else
    {
	xprintf (txts.Tant_pis);
    }
    RETURN (OK);
#undef CLEAN_UP
}


/* Add recipient to a text */
Export Success
cmd_add_recipient (String argument)
{
#define FUNCTION "add_recipient()"
    Text_no wanted_text;
    Conf_no conf_to_add_to;
    Conf_z_info add_to_conf = EMPTY_CONF_Z_INFO;

#define CLEAN_UP()   				\
    { 						\
	s_clear(&argument);			\
    }

    TOPLOOP_SETJMP();


    {
	wanted_text = reading_state.last_viewed;
	if(wanted_text==0) 
	{
	    xprintf("%s",txts.Must_read_first);
	    RETURN (FAILURE);
	}
	add_to_conf = ask_for_conf(&argument,
				   txts.What_conf_to_add);
      
	conf_to_add_to = add_to_conf.conf_no;

	if(conf_to_add_to == 0)
	    return FAILURE;

	if(kom_add_recipient(wanted_text,conf_to_add_to,0) != FAILURE)
	{
	    xprintf("%s %d %s ",
		    txts.Have_added, wanted_text, txts.parse_to_recipient);
	    print_conf_name(conf_to_add_to);
	}
	else
	    switch (kom_errno)
	    {
	    case  KOM_PERM:
		xprintf("%s %d %s ",
			txts.No_perm_to_add, wanted_text, 
			txts.parse_to_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_ALREADY_RECIPIENT:
		xprintf("%s ",txts.Already_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_RECIPIENT_LIMIT:
		xprintf("%s\n", txts.Too_many_recipients);
		break;
	    default:
		fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
			"add_recipient(%d, %d, 0) error %d",
			wanted_text, conf_to_add_to, kom_errno);
		break;
	    }
    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}


/* Copy text to another conference */
Export Success
cmd_copy_recipient (String argument)
{
#define FUNCTION "copy_recipient()"
    Text_no wanted_text;
    Conf_no conf_to_add_to;
    Conf_z_info add_to_conf = EMPTY_CONF_Z_INFO;

#define CLEAN_UP()					\
    {							\
	s_clear(&argument);				\
    }

    TOPLOOP_SETJMP();


    {
	wanted_text = reading_state.last_viewed;
	if(wanted_text==0) 
	{
	    xprintf("%s",txts.Must_read_first);
	    RETURN (FAILURE);
	}
	add_to_conf = ask_for_conf(&argument,
				   txts.What_conf_to_add);
      
	conf_to_add_to = add_to_conf.conf_no;

	if(conf_to_add_to == 0)
	    return FAILURE;

	if(kom_add_recipient(wanted_text,conf_to_add_to,1) != FAILURE)
	{
	    xprintf("%s %d %s ",
		    txts.Have_added, wanted_text, txts.parse_to_recipient);
	    print_conf_name(conf_to_add_to);
	}
	else
	    switch (kom_errno)
	    {
	    case  KOM_PERM:
		xprintf("%s %d %s ",
			txts.No_perm_to_copy, wanted_text, 
			txts.parse_to_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_ALREADY_RECIPIENT:
		xprintf("%s ",txts.Already_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_RECIPIENT_LIMIT:
		xprintf("%s\n", txts.Too_many_recipients);
		break;
	    default:
		fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
			"copy_recipient(%d, %d, 1) error %d",
			wanted_text, conf_to_add_to, kom_errno);
		break;
	    }
    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}


/* Remove recipient from a text */
Export Success
cmd_sub_recipient (String argument)
{
#define FUNCTION "sub_recipient()"
    Text_no wanted_text;
    Conf_no conf_to_sub_from;
    Conf_z_info sub_from_conf = EMPTY_CONF_Z_INFO;

#define CLEAN_UP()				\
    {						\
	s_clear(&argument); 			\
    }

    TOPLOOP_SETJMP();

    {
	wanted_text = reading_state.last_viewed;
	if(wanted_text==0)
	{
	    xprintf("%s",txts.Must_read_first);
	    RETURN (FAILURE);
	}
	sub_from_conf = ask_for_conf(&argument,
				     txts.What_conf_to_sub);
      
	conf_to_sub_from = sub_from_conf.conf_no;

	if(conf_to_sub_from == 0)
	    return FAILURE;

	if(kom_sub_recipient(wanted_text,conf_to_sub_from) != FAILURE)
	{
	    xprintf("%s %d %s ",
		    txts.Have_subed, wanted_text, txts.parse_from_recipient);
	    print_conf_name(conf_to_sub_from);
	}
	else
	    switch (kom_errno)
	    {
	    case  KOM_PERM:
		xprintf("%s %d %s ",
			txts.No_perm_to_sub, wanted_text,
			txts.parse_from_recipient);
		print_conf_name(conf_to_sub_from);
		newline();
		break;
	    case  KOM_NOT_RECIPIENT:
	        xprintf("%s",
			txts.No_such_recipient);
		newline();
		break;
	    case  KOM_NO_SUCH_TEXT:
	        xprintf("%s",
			txts.Text_doesnt_exist);
		newline();
		break;
	    default:
		fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
			"sub_recipient(%d, %d) error %d",
			wanted_text, conf_to_sub_from, kom_errno);
		break;
	    }
    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}

/* Move text */
Export Success
cmd_change_recipient (String argument)
{
#define FUNCTION "change_recipient()"
    Text_no wanted_text;
    Conf_no conf_to_add_to;
    Conf_z_info add_to_conf = EMPTY_CONF_Z_INFO;

#define CLEAN_UP()    				\
    {						\
	s_clear(&argument); 			\
    }

    TOPLOOP_SETJMP();

    {
	wanted_text = reading_state.last_viewed;
	if(wanted_text==0) 
	{
	    xprintf("%s",txts.Must_read_first);
	    RETURN (FAILURE);
	}
	add_to_conf = ask_for_conf(&argument,
				   txts.What_conf_to_add);
      
	conf_to_add_to = add_to_conf.conf_no;

	if(conf_to_add_to == 0)
	    return FAILURE;

	if(kom_add_recipient(wanted_text,conf_to_add_to,0) != FAILURE)
	{
	    if(kom_sub_recipient(wanted_text,ccn) != FAILURE)
	    {
		xprintf("%s %d %s ", 
			txts.Have_moved, wanted_text, txts.parse_to_recipient);
		print_conf_name(conf_to_add_to);
	    }
	    else
		switch (kom_errno)
		{
		case  KOM_PERM:
		    xprintf("%s %d %s ",
			    txts.No_perm_to_sub, wanted_text, 
			    txts.parse_from_recipient);
		    print_conf_name(ccn);
		    break;
		case  KOM_NOT_RECIPIENT:
		    xprintf("%s\n",txts.Must_be_in_conf);
		    break;
		default:
		    fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
			    "change_recipient(sub, %d, %d) error %d",
			    wanted_text, ccn, kom_errno);
		    break;

		}
	}
	else
	    switch (kom_errno)
	    {
	    case  KOM_PERM:
		xprintf("%s %d %s ",
			txts.No_perm_to_add, wanted_text,
			txts.parse_to_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_ALREADY_RECIPIENT:
		xprintf("%s ",txts.Already_recipient);
		print_conf_name(conf_to_add_to);
		break;
	    case  KOM_RECIPIENT_LIMIT:
		xprintf("%s\n", txts.Too_many_recipients);
		break;
	    default:
		fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
			"change_recipient(add, %d, %d, 0) error %d",
			wanted_text, conf_to_add_to, kom_errno);
		break;
	    }
    }

    RETURN (OK);
#undef CLEAN_UP
#undef FUNCTION
}
