/*
 *  error.c  --  routines for handling errors and displaying error
 *		 messages in the client.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */

#include <config.h>
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdarg.h>
#include <sys/types.h>

#include <kom-types.h>
#include <kom-errno.h>

#include "xoutput.h"
#include "texts.h"
#include "internal.h"	/* Sigh. */

#include "error.h"

#define	Export


#define errprint(msg)	write (2, msg, strlen (msg))



/*
 *  Inform the user about some error he or she did.
 *  BUG: This function is probably under-used...
 */
extern  void
inform_user (Type_of_error	  condition,
	     Conf_z_info_list	* confs)

{
#define FUNCTION "inform_user()"

    switch (condition)
    {
    case  ambigous_or_no_conf:
	if (confs == NULL  ||  confs->no_of_confs == 0)
	    xprintf ("%s.\n", txts.No_conf_with_that_name);
	else
	{
	    int i;

	    xprintf ("%s. %s:\n",
		    txts.Conf_name_ambigous,
		    txts.You_might_mean_one_of_these);
	    for (i = 0; i < confs->no_of_confs; i++)
	    {
		print_conf_name (confs->confs[i].conf_no);
		newline();
	    }
	}
	break;

#ifndef COMPILE_CHECKS
    default:
	fatal1 (CLIENT_SHOULDNT_HAPPEN, "Bad type of user mistake");
	break;
#endif
    }
#undef FUNCTION
}




/*
 *  Prints out a message (to stderr) apropriate for the error ERROR_NO.
 */
Export  void
client_error (Kom_err  error_no)

{
#define FUNCTION "client_error()"
    switch (error_no)
    {
    case  KOM_NO_ERROR:
    case  KOM_NOT_IMPL:
    case  KOM_OBSOLETE:
    case  KOM_PWD:
    case  KOM_LONG_STR:
    case  KOM_LOGIN:
    case  KOM_LOGIN_DISALLOWED:
    case  KOM_CONF_ZERO:
    case  KOM_UNDEF_CONF:
    case  KOM_UNDEF_PERS:
    case  KOM_ACCESS:
    case  KOM_PERM:
    case  KOM_NOT_MEMBER:
    case  KOM_NO_SUCH_TEXT:
    case  KOM_NO_SUCH_LOCAL_TEXT:
    case  KOM_BAD_NAME:
    case  KOM_CONF_EXISTS:
    case  KOM_PERS_EXISTS:
    case  KOM_SECRET_PUBLIC:
    case  KOM_LETTER_BOX:
    case  KOM_LDB_ERR:
    case  KOM_ILL_MISC:
    case  KOM_ILLEGAL_INFO_TYPE:
    case  KOM_ALREADY_RECIPIENT:
    case  KOM_NOT_RECIPIENT:
    case  KOM_RECIPIENT_LIMIT:
    case  KOM_COMM_LIMIT:
    case  KOM_FOOT_LIMIT:
    case  KOM_NOT_AUTHOR:
    case  KOM_NO_CONNECT:
    case  KOM_OUT_OF_MEMORY:
    case  KOM_SERVER_IS_CRAZY:
    case  KOM_CLIENT_IS_CRAZY:
	kom_perror();
	break;

#ifndef COMPILE_CHECKS
    default:
	fatal1 (CLIENT_SHOULDNT_HAPPEN, "client_error()");
	break;
#endif
    }
#undef FUNCTION
}



Export  void
fatal_client_error (Fatal_client_error	  error,
		    const char		* function_name,
		    const char		* filename,
		    const int		  lineno,
		    const char		* descrfmt, ...)

{
#define FUNCTION "fatal_client_error"
    va_list	  argp;


    errprint (txts.fatal_error_occured);

    switch (error)
    {
    case  CLIENT_OUT_OF_MEMORY:
	errprint (txts.out_of_memory);
	break;

    case  CLIENT_SERVER_ERROR:
	if (kom_errno == KOM_NO_CONNECT) {
	    fprintf (stderr, "%s.\n",
		     txts.Server_doesnt_respond);
	    fprintf (stderr, "kom_errno = %d\n", kom_errno);
	    break;
	}
	fprintf (stderr, "%s: %s\n",
		 txts.Fatal_error,
		 txts.Unexpected_server_error);
	fprintf (stderr, "kom_errno = %d\n", kom_errno);
	break;

    case  CLIENT_CONNECTION_LOST:
	fprintf (stderr, "%s: %s\n",
		 txts.Fatal_error,
		 txts.Server_connection_lost);
	fprintf (stderr, "kom_errno = %d\n", kom_errno);
	break;

    case  CLIENT_UNEXPECTED_SERVER_ERROR:
	fprintf (stderr, "%s: %s\n",
		 txts.Fatal_error,
		 txts.Unexpected_server_error);
	fprintf (stderr, "kom_errno = %d\n", kom_errno);
	break;

    case  CLIENT_SITUATION_NOT_HANDLED:
	errprint (txts.Situation_not_handled);
	fprintf (stderr, "\nkom_errno = %d\n", kom_errno);
	break;

    case  CLIENT_SHOULDNT_HAPPEN:
	fprintf (stderr, "%s %s: %d\n",
		 txts.Internal_error_in_function,
		 function_name,
		 lineno);
	break;

    case  CLIENT_ERROR_TOO_boring:
	fprintf (stderr, "\a\a\aAaarrrggghhh!!! This error isn't handled!\n");
	fprintf (stderr, "kom_errno = %d\n", kom_errno);
	break;

#ifndef COMPILE_CHECKS
    default:
	fatal1 (CLIENT_SHOULDNT_HAPPEN, "fatal");
	break;
#endif
    }

    errprint (txts.debug_info_follows);
    errprint ("Client version: " VERSION "\n");
    errprint ("Current function: ");
    errprint (function_name);
    errprint ("\tFile: ");
    errprint (filename);
    {
	char  buf[12];
	
	sprintf(buf, ":%d", lineno);
	errprint (buf);
    }
    errprint ("\n");
    va_start (argp, descrfmt);
    vfprintf (stderr, descrfmt, argp);
    va_end (argp);
    errprint ("\n");
    abort();
#undef FUNCTION
}



/*
 *  Print a message telling the user that a bug in the LysKOM
 *  server has been detected, the symtoms of the bug (BUG_TEXT),
 *  in what function in the client the bug was detected
 *  (FUNCTION) and some more information that might be used for
 *  post-mortem debugging.
 */
Export  void
server_bug (const char	* bug_text,
	    const char	* function,
	    bug_info	  extra_info)

{
    xprintf ("Oops! I have detected a bug in the LysKOM server.\n"
	     "Symtoms: %s\n"
	     "Function where the bug was detected: %s\n"
	     "Extra debug-info: %ul\n",
	     bug_text, function, (unsigned long) extra_info);
    xprintf ("\nI will dump core now. Please save the core file, and\n"
	     "tell someone who is responsible for this LysKOM\n"
	     "installation about what happened, so that the bug\n"
	     "can be fixed.  Thanks.\n");
    xprintf ("\nThe person you should contact is linus@lysator.liu.se\n");

    abort();
}
