/*
 *  client/main.c  --  LysKOM client, main program
 *
 *
 *  Copyright (C) 1990  Lysator Computer Club
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  E-mail:	Bellman@Lysator.LiU.SE
 */


#include <config.h>
#include <stdio.h>
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef __vax__
#define __SIZE_T included from stdlib.h
#define __WCHAR_T included from stdlib.h
#endif
#include <kom-types.h>
#include <zmalloc.h>
#include <s-string.h>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include <services.h>

/* Must be included before "read-line.h" on a vax because it includes <sgtty.h>
 * and that does not work after inclusion of <sys/ioctl.h> because of
 * stupid include-file construction.
 */
#include "xoutput.h"

#include "error.h"
#include "read-line.h"
#include "texts.h"
#include "commands.h"
#include "cmds.h"
#include "internal.h"
#include "cache.h"
#include "edit.h"

#include "defaults.h"

char		* server_name;
int		  portno;
Info		  serv_info;
char		* client_name;

int
main (int	   argc,
      char	** argv)

{
#define FUNCTION "main()"

    s_set_storage_management (zmalloc, zrealloc, zfree);
    kom_set_storage_management (zmalloc, zrealloc, zfree);
#ifdef HAVE_LOCALE_H
    setlocale(LC_ALL, "");
#endif
    init_messages();	/* Initialize the 'txts' struct */
    init_cmd_list();	/* Initialize the list of commands */
    init_terminal (stdin, stdout);
    init_cache();

    client_name = argv[0];
    if (!client_name)
	client_name = "(unknown)";

    server_name = DEFAULT_SERVER;
    portno = DEFAULT_PORTNO;

    /* Aronsson was here: 7 bit character size conversion for terminals */
    /* For now, argument parsing is very primitive. */
    /* Perhaps we should consider GNU's getopt. */
    while (argc > 1)
    {
        if (0 == strcmp (argv[1], "--iso"))
	    charset=0;
	else if (0 == strcmp (argv[1], "--pc8"))
	    charset=1;
	else if (0 == strcmp (argv[1], "--sw7"))
	    charset=2;
	else if (0 == strcmp (argv[1], "--mac"))
	    charset=3;
	else if (0 == strcmp (argv[1], "--tomradavslutar"))
	    tomrad_avslutar = 1;
	else if (argv[1][0] != '-')
	{
	    char * p;

	    /* Don't use strdup here. It does not exist on ULTRIX */
	    server_name = (char *) zmalloc(strlen(argv[1]) + 1);
	    strcpy(server_name, argv[1]);

	    if ((p = strchr(server_name, ':')) != NULL)
	    {
		portno = atoi(p + 1);
		*p = '\0';
	    }
	} else {
	    xprintf ("Usage: [--iso] [--pc8] [--mac] [--sw7] [--tomradavslutar] [server]\n");
	    restore_terminal (stdin, stdout);
	    exit(1);
	}
	
	argc --;
	argv ++;
    }
    /* Here argv and argc are irreparably destroyed. */


    if (kom_open_connection(server_name, portno) == FAILURE)
    {
	xprintf (txts.no_contact_with_server, server_name);
	restore_terminal (stdin, stdout);
	exit(1);
    }

    /*  OK, connection established  */
    kom_get_info (&serv_info);	/* BUG! Retval not checked */

    /*  Greet user  */
    xprintf (txts.greet_user /* , serv_info.server_id  Vad {r detta? */,
	     server_name,
	     serv_info.version,
	     VERSION);
    xprintf ("\n");
    if (serv_info.motd_of_lyskom)
        show_text_no (serv_info.motd_of_lyskom);

    /*  Log in to server  */
    cmd_start_anew (EMPTY_STRING);
    
    /*  And run the command loop  */
    command_loop ();

    /* Restore the terminal to the old settings */
    restore_terminal (stdin, stdout);

    return OK;	/*  LysKOM can *never* be wrong! */
#undef FUNCTION
}

